/* PatchModuleName - modify module name to help dumb utilities to the right thing
   Some utilities treat module names as case sensitive or get confused if an
   EXE and a DLL have the same module name.
   Dumb utilities include pmdf dllrname and ida.
   Ida is more antique than dumb since it is a DOS app.
   PMDF gets confused because it uses the module name to associate the .sym file.
   The module name can be changed to uppercase or lowercase.
   The last letter of an EXE module name can be changed to X avoid module
   name confusion.

   Copyright (c) 2010-2016 Steven Levine and Associates, Inc.
   All rights reserved.

   This program is free software licensed under the terms of the GNU
   General Public License.  The GPL Software License can be found in
   gnugpl2.txt or at http://www.gnu.org/licenses/licenses.html#GPL

   2010-02-08 SHL Baseline
   2012-09-27 SHL Sync with templates
   2012-09-27 SHL Make smarter
   2014-06-12 SHL Sync with templates
   2015-11-07 SHL Sync with templates
   2015-11-07 SHL Drop unneeded setlocal
   2016-03-23 SHL Support case edits
*/

signal on Error
signal on FAILURE name Error
signal on Halt
signal on NOTREADY name Error
signal on NOVALUE name Error
signal on SYNTAX name Error

Globals = 'gArgList. gCmdName gDbgLvl gEnv gFileName gMakeXName',
	  'gNewCase gTesting gTmpDir gVerbose gVersion'

gVersion = '0.1'

call Initialize

Main:

  parse arg cmdLine
  call ScanArgs cmdLine
  drop cmdLine

  call DoWildArgs

  exit

/* end main */

/*=== DoWildArgs() Process wildcarded argument list ===*/

DoWildArgs: procedure expose (Globals)

  do argNum = 1 to gArgList.0
    wildCard = gArgList.argNum
    call DbgMsg 'Searching for' wildCard

    /* Format options F: files D: directories S: recurse */
    /* MM-DD-YY HH:MM Size ADHRS Name (default) */
    /* L: MM-DD-YYYY HH:MM Size ADHRS Name */
    /* O: Name only */
    /* T: YY/MM/DD/HH/MM Size ADHRS Name */
    /* TL: YYYY-MM-DD HH:MM:SS Size ADHRS Name */
    call SysFileTree wildCard, 'fileList', 'FL'
    if RESULT \= 0 then
      call Die 'SysFileTree' wildCard 'failed'
    do fileNum = 1 to FileList.0
      parse var FileList.fileNum fileDate fileTime fileBytes fileAttrib fileName
      fileName = strip(fileName)
      call DoOneArg fileName
    end /* fileNum */
  end /* argNum */

  return

/* end DoWildArgs */

/*=== DoOneArg(pathName) Patch one file ===*/

DoOneArg: procedure expose (Globals)

  parse arg pathName

  say

  i = lastpos('\', pathName)
  fileName = substr(pathName, i + 1)
  i = lastpos('.', fileName)
  if i = 0 then
    baseName = fileName
  else
    baseName = substr(fileName, 1, i - 1)

  ext = translate(substr(fileName, i + 1))

  call DbgMsg 'Base name is' baseName

  buf = charin(pathName, 1, chars(pathName))	/* Gulp */
  call stream pathName, 'C', 'CLOSE'

  lxOffset = 0				/* 0 relative */
  if left(buf, 2) == 'MZ' then do
    /* Have MZ header, find LX header location */
    flag = substr(buf, 24 + 1, 1)
    if flag \== '40'x then
      call Die fileName 'missing new header flags at 0x18'
    /* Get offset to new exe header */
    lxOffset = c2d(reverse(substr(buf, x2d('3c') + 1, 2)))
  end
  if substr(buf, lxOffset + 1, 2) \== 'LX' then
    call Die fileName 'does not appear to be an LX execuatable'

  offset = lxOffset + x2d('58')		/* Offset to resident name table offset in header */

  call DbgMsg 'Resident name table at file offset' d2x(offset) '('offset')'
  /* Get offset of resident names table */
  x = c2x(reverse(substr(buf, offset + 1, 2)))
  d = x2d(x)
  offset = lxOffset + d

  /* Module name is first name listed */
  call DbgMsg 'Module name at file offset' d2x(offset) '('offset')'

  x = c2x(substr(buf, offset + 1, 1))	/* Get byte count */
  len = x2d(x)

  call DbgMsg 'Module name length is' len 'bytes'

  modName = substr(buf, offset + 2, len)

  say 'Current module name is' modName

  changed = 0

  /* Edit module name as requested */

  newName = modName

  select
  when gNewCase = 'lower' then
    newName = ToLower(newName)
  when gNewCase = 'upper' then
    newName = translate(newName)
  otherwise
    nop
  end

  if gMakeXName then do
    c = right(newName, 1)
    if translate(c) == 'X' then
      call Die 'Can not patch module name that ends with the letter X'
    /* Match case */
    if translate(c) == c then
      c = 'X'
    else
      c = 'x'
    newName = left(newName, len - 1) || c
  end

  if modName == newName then
    call Warn 'No changes to module name - file will not be modified'
  else do
    call DbgMsg 'New module name will be' newName

    buf = overlay(newName, buf, offset + 2)

    s = MakePath('', gTmpDir, fileName, '')
    /* Backup only one time */
    if \ IsFile(s) then do
      call DbgMsg 'Backing up original to' s
      'copy' pathName s
    end
    call charout pathName, buf, 1
    call stream pathName, 'C', 'CLOSE'

    say 'Module name changed to' newName
  end

  return

/* end DoOneArg */

/*=== Initialize() Initialize globals ===*/

Initialize: procedure expose (Globals)
  call GetCmdName
  call LoadRexxUtil
  gEnv = 'OS2ENVIRONMENT'
  call GetTmpDir			/* Used for backup files */
  return

/* end Initialize */

/*=== ScanArgsInit() ScanArgs initialization exit routine ===*/

ScanArgsInit: procedure expose (Globals) cmdTail swCtl keepQuoted
  /* Preset defaults */
  gDbgLvl = 0				/* Display debug messages */
  gNewCase = ''				/* Change case to upper, lower or leave unchanged */
  gMakeXName = 0			/* Change last char of module name to X */
  gArgList.0 = 0			/* Reset arg count */
  return

/* end ScanArgsInit */

/*=== ScanArgsSwitch() ScanArgs switch option exit routine ===*/

ScanArgsSwitch: procedure expose (Globals) curSw curSwArg

  select
  when curSw == 'd' then
    gDbgLvl = gDbgLvl + 1
  when curSw == 'l' then
    gNewCase = 'lower'
  when curSw == 'h' | curSw == '?' then
    call ScanArgsHelp
  when curSw == 'u' then
    gNewCase = 'upper'
  when curSw == 'V' then do
    say gCmdName gVersion
    exit
  end
  when curSw == 'x' then
    gMakeXName = 1
  otherwise
    call ScanArgsUsage 'switch '''curSw''' unexpected'
  end /* select */

  return

/* end ScanArgsSwitch */

/*=== ScanArgsArg() ScanArgs argument option exit routine ===*/

ScanArgsArg: procedure expose (Globals) curArg

  if \ IsFile(curArg) then
    call ScanArgsUsage curArg 'not found'

  i = lastpos('\', curArg)
  if i = 0 then
    i = 1
  i = pos('.', curArg, 1)
  if i = 0 then
    call ScanArgsUsage curArg 'must include extension'

  ext = translate(substr(curArg, i + 1))

  if ext \== 'EXE' & ext \== 'DLL' then
    call ScanArgsUsage curArg 'file extensions must exe or dll'

  if gMakeXName & ext == 'DLL' then
    call ScanArgsUsage '-x not allowed for DLLs'

  i = gArgList.0 + 1
  gArgList.i = curArg
  gArgList.0 = i

  return

/* end ScanArgsArg */

/*=== ScanArgsTerm() ScanArgs scan end exit routine ===*/

ScanArgsTerm: procedure expose (Globals)
  if gArgList.0 = 0 then
    call ScanArgsUsage
  return

/* end ScanArgsTerm */

/*=== ScanArgsHelp() Display ScanArgs usage help exit routine ===*/

ScanArgsHelp:
  say
  say 'Patch executable file module names to allow dumb utilities to the right thing.'
  say 'Can change module name to upper or lower case.'
  say 'Dllrname requires uppercase module names.'
  say 'Can change last letter of EXE module name to X to avoid confusion with a'
  say 'same-named DLL.'
  say 'Pmdf requires a unique module name to associate the correct SYM file'
  say 'with an executeable.'
  say 'Files are edited in place.  The unedited original is copied to %TMP.'
  say
  say 'Usage:' gCmdName '[-d] [-h] [-l] [-u] [-V] [-x] [-?] filename...'
  say
  say '  -d        Enable debug logic, repeat for more verbosity'
  say '  -h -?     Display this message'
  say '  -l        Change module name to lower case'
  say '  -u        Change module name to upper case'
  say '  -V        Display version number and quit'
  say '  -x        Change last letter of module name to x'
  say
  say '  filename  Executable file name (DLL or EXE) - wildcards allowed'
  exit 255

/* end ScanArgsHelp */

/*=== ScanArgsUsage(message) Report Scanargs usage error exit routine ===*/

ScanArgsUsage:
  parse arg msg
  say
  if msg \== '' then
    say msg
  say 'Usage:' gCmdName '[-d] [-h] [-l] [-u] [-V] [-x] [-?] filename...'
  exit 255

/* end ScanArgsUsage */

/*==============================================================================*/
/*=== SkelRexxFunc standards - Delete unused - Move modified above this mark ===*/
/*==============================================================================*/

/*=== DbgMsg([level, ]message,...) Write multi-line message to STDERR if debugging ===*/

DbgMsg: procedure expose (Globals)
  level = arg(1)
  if dataType(level, 'W') then
    start = 2
  else do
    level = 1
    start = 1
  end
  if level <= gDbgLvl then do
    do i = start to arg()
      msg = arg(i)
      if msg \== '' then
	msg = ' *' msg
      call lineout 'STDERR', msg
    end
  end
  return

/* end DbgMsg */

/*=== IsFile(fileSpec) return true if arg is file ===*/

IsFile: procedure expose (Globals)
  parse arg fileSpec
  if fileSpec == '' then
    yes = 0
  else do
    call SysFileTree fileSpec, 'fileList', 'F'
    if RESULT \= 0 then
      call Die 'IsFile' wildCard 'failed'
    /* Assume caller knows if arg contains wildcards */
    yes = fileList.0 \= 0
  end
  return yes

/* end IsFile */

/*=== MakePath(drv, dir, name, ext) Make pathname from parts ===*/

MakePath: procedure

  /* All parts optional - code guesses what caller means
     MakePath(path, filename) will work
     MakePath(path, name, ext) will not work
     Automatically converts unix slashes to dos slashes
   */

  parse arg drv, dir, name, ext

  /* If 2 args, assume dir and name */
  if name == '' & ext == '' then do
    name = dir
    dir = drv
    drv = ''
  end

  if drv == '' then
    path = ''				/* Drive omitted */
  else do
    drv = translate(drv, '\', '/')	/* Ensure DOS */
    path = drv
    /* If just one character assume drive letter */
    if length(drv) = 1 then
      path = path':'
    /* If leading \\ assume UNC */
    else if left(drv, 2) == '\\' & right(drv, 1) \== '\' then
      path = path'\'
    /* If 3 args and drv longer than 2 and no trailing \ on dir assume given dir, name, ext */
    else if ext == '' & right(dir, 1) \= '\' & length(drv) > 2 then do
      ext = name
      name = dir
      dir = ''
      drv = ''
    end
    /* Otherwise figure out later */
  end

  if dir \== '' then do
    dir = translate(dir, '\', '/')	/* Ensure DOS */
    c = right(path, 1)
    c2 = left(dir, 1)
    if path \== '' & c \== ':' & c \== '\' & c2 \== '\' then
      path = path || '\' || dir
    else if c == '\' & c2 == '\' then
      path = path || substr(dir, 2)
    else
      path = path || dir
  end

  if name \== '' then do
    c = right(path, 1)
    if path \== '' & c \= '\' & c \== ':' then
      path = path || '\' || name
    else
      path = path || name
  end

  if ext \== '' then do
    if left(ext, 1) \== '.' then
      path = path || '.' || ext
    else
      path = path || ext
  end

  return path

/* end MakePath */

/*=== ToLower(s) Convert to lower case ===*/

ToLower: procedure
  parse arg s
  return translate(s, xrange('a', 'z'), xrange('A', 'Z'))

/* end ToLower */

/*=== Warn(message,...) Write multi-line warning message to STDERR ===*/

Warn: procedure
  do i = 1 to arg()
    msg = arg(i)
    call lineout 'STDERR', msg
  end
  return

/* end Warn */

/*==========================================================================*/
/*=== SkelRexx standards - Delete unused - Move modified above this mark ===*/
/*==========================================================================*/

/*=== Error() Report ERROR etc. to STDOUT and trace and exit or return if called ===*/

Error:
  say
  parse source . . cmd
  say condition('C') 'signaled at line' SIGL 'of' cmd'.'
  if condition('D') \= '' then say 'REXX reason =' condition('D')'.'
  if condition('C') == 'SYNTAX' & symbol('RC') == 'VAR' then
    say 'REXX error =' RC '-' errortext(RC)'.'
  else if symbol('RC') == 'VAR' then
    say 'RC =' RC'.'
  say 'Source =' sourceline(SIGL)

  if condition('I') \== 'CALL' | condition('C') == 'NOVALUE' | condition('C') == 'SYNTAX' then do
    trace '?A'
    say 'Enter REXX commands to debug failure.  Press enter to exit script.'
    call SysSleep 2
    if symbol('RC') == 'VAR' then exit RC; else exit 255
  end

  return

/* end Error */

/*=== Die([message,...]) Write multi-line error message to stderr and die ===*/

Die:
  call lineout 'STDERR', ''
  do i = 1 to arg()
    call lineout 'STDERR', arg(i)
  end
  call lineout 'STDERR', gCmdName 'aborting at script line' SIGL
  call Beep 200, 300
  call SysSleep 2
  exit 254

/* end Die */

/*=== GetCmdName() Get short script name and set gCmdName ===*/

GetCmdName: procedure expose (Globals)
  parse source . . cmd
  cmd = filespec('N', cmd)		/* Chop path */
  c = lastpos('.', cmd)
  if c > 1 then
    cmd = left(cmd, c - 1)		/* Chop extension */
  gCmdName = translate(cmd, xrange('a', 'z'), xrange('A', 'Z'))	/* Lowercase */
  return

/* end GetCmdName */

/*=== GetTmpDir() Get TMP dir name with trailing backslash, set gTmpDir ===*/

GetTmpDir: procedure expose (Globals)
  s = value('TMP',,gEnv)
  if s \= '' & right(s, 1) \= ':' & right(s, 1) \== '\' then
    s = s'\'				/* Stuff backslash */
  gTmpDir = s
  return

/* end GetTmpDir */

/*=== Halt() Report HALT condition and exit ===*/

Halt:
  say
  parse source . . cmd
  say condition('C') 'signaled at' cmd 'line' SIGL'.'
  say 'Source = ' sourceline(SIGL)
  say 'Sleeping for 2 seconds...'
  call SysSleep 2
  exit 253

/* end Halt */

/*=== LoadRexxUtil() Load RexxUtil functions ===*/

LoadRexxUtil:
  if RxFuncQuery('SysLoadFuncs') then do
    call RxFuncAdd 'SysLoadFuncs', 'REXXUTIL', 'SysLoadFuncs'
    if RESULT then
      call Die 'Cannot load SysLoadFuncs'
    call SysLoadFuncs
  end
  return

/* end LoadRexxUtil */

/*=== ScanArgs(cmdLine) Scan command line ===*/

ScanArgs: procedure expose (Globals)

  /* Calls user exits to process arguments and switches */

  parse arg cmdTail
  cmdTail = strip(cmdTail)

  call ScanArgsInit

  /* Ensure optional settings initialized */
  if symbol('SWCTL') \== 'VAR' then
    swCtl = ''				/* Switches that take args, append ? if optional */
  if symbol('KEEPQUOTED') \== 'VAR' then
    keepQuoted = 0			/* Set to 1 to keep arguments quoted */

  /* Scan */
  curArg = ''				/* Current arg string */
  curSwList = ''			/* Current switch list */
  /* curSwArg = '' */			/* Current switch argument, if needed */
  noMoreSw = 0				/* End of switches */

  do while cmdTail \== '' | curArg \== '' | curSwList \== ''

    if curArg == '' then do
      /* Buffer empty, refill */
      qChar = left(cmdTail, 1)		/* Remember quote */
      if \ verify(qChar,'''"', 'M') then
	parse var cmdTail curArg cmdTail	/* Not quoted */
      else do
	/* Arg is quoted */
	curArg = ''
	do forever
	  /* Parse dropping quotes */
	  parse var cmdTail (qChar)quotedPart(qChar) cmdTail
	  curArg = curArg || quotedPart
	  /* Check for escaped quote within quoted string (i.e. "" or '') */
	  if left(cmdTail, 1) \== qChar then
	    leave			/* No, done */
	  curArg = curArg || qChar	/* Append quote */
	  if keepQuoted then
	    curArg = curArg || qChar	/* Append escaped quote */
	  parse var cmdTail (qChar) cmdTail
	end /* do */
	if keepQuoted then
	  curArg = qChar || curArg || qChar	/* requote */
      end /* if quoted */
    end

    /* If switch buffer empty, refill */
    if curSwList == '' then do
      if left(curArg, 1) == '-' & curArg \== '-' then do
	if noMoreSw then
	  call ScanArgsUsage 'switch '''curArg''' unexpected'
	else if curArg == '--' then
	  noMoreSw = 1
	else do
	  curSwList = substr(curArg, 2)	/* Remember switch string */
	  curArg = ''			/* Mark empty */
	  iterate			/* Refill arg buffer */
	end
	parse var cmdTail curArg cmdTail
      end
    end

    /* If switch in progress */
    if curSwList \== '' then do
      curSw = left(curSwList, 1)	/* Next switch */
      curSwList = substr(curSwList, 2)	/* Drop from pending */
      /* Check switch allows argument, avoid matching ? */
      if pos(curSw, translate(swCtl,,'?')) \= 0 then do
	if curSwList \== '' then do
	  curSwArg = curSwList		/* Use rest of switch string for switch argument */
	  curSwList = ''
	end
	else if curArg \== '' & left(curArg, 1) \== '-' then do
	  curSwArg = curArg		/* Arg string is switch argument */
	  curArg = ''			/* Mark arg string empty */
	end
	else if pos(curSw'?', swCtl) = 0 then
	  call ScanArgsUsage 'Switch' curSw 'requires argument'
	else
	  curSwArg = ''			/* Optional arg omitted */
      end

      call ScanArgsSwitch		/* Passing curSw and curSwArg */
      drop curSwArg			/* Must be used by now */
    end /* if switch */

    /* If arg */
    else if curArg \== '' then do
      noMoreSw = 1
      call ScanArgsArg			/* Passing curArg */
      curArg = ''
    end

  end /* while not done */

  call ScanArgsTerm

  return

/* end ScanArgs */

/* The end */

